unit WEBPFormat;

{$IFDEF FPC}
  {$MODE delphiunicode}
{$ENDIF}

interface

uses Classes, Windows, SysUtils;

type
  TWebpCompression = (webpUnknown, webpLossless, webpLossy);

  { TWEBPInfo }

  TWEBPInfo = record
    Valid: Boolean;
    Width, Height, FrameCount: Integer;
    Extended, ICC, Exif, Alpha, XMP, Animation, FirstFrameAlpha: Boolean;
    Duration: Integer;
    Compression: TWebpCompression;
    function BitDepth: Integer;
  end;

function ReadWEBPHeader(const Stream: TStream; var WEBPInfo: TWEBPInfo): Boolean;

implementation

{
https://theroadtodelphi.com/2009/10/21/bits-manipulation-functions-using-delphi/
http://www.delphigroups.info/2/21/175350.html
}

const
  {Here we count the bit 0 to 7}
  mask:array[0..7] of byte = (1,2,4,8,16,32,64,128);

function BitTest(a, bit: byte): boolean;
begin
  BitTest := a and mask[bit] <> 0;
end;

procedure BitSet(var a: byte; bit: byte);
begin
  a := a or mask[bit];
end;

procedure BitClear(var a: byte; bit: byte);
begin
  a := a and not mask[bit];
end;

procedure BitCopy(source, sourcebit: byte; var dest: byte; destbit: byte);
begin
  if BitTest(source, sourcebit) then
    BitSet(dest, destbit)
  else
    BitClear(dest, destbit);
end;

//get if a particular bit is 1
function Get_a_Bit(const aValue: Cardinal; const Bit: Byte): Boolean;
begin
  Result := (aValue and (1 shl Bit)) <> 0;
end;

//set a particular bit as 1
function Set_a_Bit(const aValue: Cardinal; const Bit: Byte): Cardinal;
begin
  Result := aValue or (1 shl Bit);
end;

//set a particular bit as 0
function Clear_a_Bit(const aValue: Cardinal; const Bit: Byte): Cardinal;
begin
  Result := aValue and not (1 shl Bit);
end;

//Enable o disable a bit
function Enable_a_Bit(const aValue: Cardinal; const Bit: Byte; const Flag: Boolean): Cardinal;
begin
  Result := (aValue or (1 shl Bit)) xor (Integer(not Flag) shl Bit);
end;

function ByteToBin(aByte: byte): string;
const
  c10: array [Boolean] of Char = ('0', '1');
var
  eLoop1: byte;
begin
  SetLength(Result, 8);
  for eLoop1 := 7 downto 0 do
    Result[8 - eLoop1] := c10[(aByte and (1 shl eLoop1)) <> 0];
end;


type
  TRiffChunk = packed record
    Name: array [0..3] of AnsiChar;
    Size: Cardinal; // The size in bytes starting at offset 8
  end;

  TWebpHeader = packed record
    RIFF: array [0..3] of AnsiChar; // 'RIFF'
    Size: Cardinal; // The size of the file in bytes starting at offset 8
    WEBP: array [0..3] of AnsiChar; // 'WEBP'
    Format: array [0..3] of AnsiChar; // 'VP8 ' - simple, 'VP8L' - lossless, 'VP8X' - extended
  end;

  TWebpExtendedChunk = packed record
    ChunkSize: Cardinal;
    Includes: Byte;
    Dummy1, Dummy2, Dummy3: Byte;
    CanvasWidth, CanvasHeight: array [0..2] of Byte; // real size is CW+1, CH+1
  end;

  TWebpLosslessChunk = packed record
    ChunkSize: Cardinal;
    Signature: Byte;
    Width, Height: array [0..1] of Byte; // real size is W+1, H+1
  end;

  TWebpSimpleChunk = packed record
    ChunkSize: Cardinal;
    FrameTag: array [0..2] of Byte;
    StartCodeByte: array [0..2] of Byte;
    Width, Height: array [0..1] of Byte; // as is, no +1!
  end;

  TWebpANMFChunk = packed record
    ChunkSize: Cardinal;
    A1, A2, A3,
    Duaration: Cardinal;
  end;

function FixOdd(Value: Cardinal): Cardinal; inline;
begin
  if Odd(Value) then
    Result := Value + 1
  else
    Result := Value;
end;

procedure ReadNextChunks(const Stream: TStream; WebpSize: Cardinal; var WEBPInfo: TWEBPInfo);
var
  Chunk, TempChunk: TRiffChunk;
  BytesRead: Integer;
  Animated: Boolean;
  FrameCount, Counter: Integer;
  APos, FirstANMFPos: Int64;
  WA: TWebpANMFChunk;
  WL: TWebpLosslessChunk;
  I: Integer;
begin
  Animated := False;
  FrameCount := 0;
  Counter := 0;
  APos := Stream.Position;

  repeat
    if Stream.Position > (APos + 8 + WebpSize) then
      Break;
    BytesRead := Stream.Read(Chunk, SizeOf(Chunk));
    if BytesRead <> SizeOf(Chunk) then
      Break;
    if (Chunk.Size + Stream.Position) > ((Stream.Size - APos + 8) + WebpSize)  then
      Break;

    if Chunk.Name = 'VP8 ' then
    begin
      if WEBPInfo.Compression = webpUnknown then
        WEBPInfo.Compression := webpLossy;
    end else if Chunk.Name = 'VP8L' then
    begin
      Inc(Counter);

      if WEBPInfo.Compression = webpUnknown then
        WEBPInfo.Compression := webpLossless;

      APos := Stream.Position;
      Stream.Position := Stream.Position - 4;

      BytesRead := Stream.Read(WL, SizeOf(WL));
      if BytesRead <> SizeOf(WL) then
        Break;
      if BitTest(WL.Height[1], 4) then
      begin
        WEBPInfo.Alpha := True;
        if Counter = 1 then
          WEBPInfo.FirstFrameAlpha := True;
      end else begin
        if Counter = 1 then
          WEBPInfo.FirstFrameAlpha := False;
      end;

      Stream.Position := APos;
    end else if Chunk.Name = 'ANIM' then
      Animated := True
    else if Chunk.Name = 'ANMF' then
    begin
      Inc(FrameCount);

      if FrameCount = 1 then // first frame
      begin
        FirstANMFPos := Stream.Position;
        WEBPInfo.FirstFrameAlpha := False;

        Stream.Position := Stream.Position - 4;
        BytesRead := Stream.Read(WA, SizeOf(WA));
        if BytesRead <> SizeOf(WA) then
          Break;

        for I := 24 to 31 do
          WA.Duaration := Clear_a_Bit(WA.Duaration, I);
        WEBPInfo.Duration := WA.Duaration;

        BytesRead := Stream.Read(TempChunk, SizeOf(TempChunk));
        if BytesRead <> SizeOf(TempChunk) then
          Break;

        if TempChunk.Name = 'ALPH' then
        begin
          //if not WEBPInfo.Alpha then
          begin
            WEBPInfo.Alpha := True;
            WEBPInfo.FirstFrameAlpha := True;
          end;
          Stream.Position := Stream.Position + FixOdd(TempChunk.Size);
          BytesRead := Stream.Read(TempChunk, SizeOf(TempChunk));
          if BytesRead <> SizeOf(TempChunk) then
            Break;
        end;

        if TempChunk.Name = 'VP8 ' then
        begin
          if WEBPInfo.Compression = webpUnknown then
            WEBPInfo.Compression := webpLossy;
        end else if TempChunk.Name = 'VP8L' then
        begin
          if WEBPInfo.Compression = webpUnknown then
            WEBPInfo.Compression := webpLossless;

          Stream.Position := Stream.Position - 4;
          BytesRead := Stream.Read(WL, SizeOf(WL));
          if BytesRead <> SizeOf(WL) then
            Break;
          if BitTest(WL.Height[1], 4) then
          begin
            WEBPInfo.Alpha := True;
            WEBPInfo.FirstFrameAlpha := True;
          end else begin
            WEBPInfo.FirstFrameAlpha := False;
          end;
        end;

        Stream.Position := FirstANMFPos;
      end;
    end;

    Stream.Position := Stream.Position + FixOdd(Chunk.Size);

  until False;


  if Animated then
    WEBPInfo.Animation := True;
  if FrameCount > 0 then
    WEBPInfo.FrameCount := FrameCount
  else
    WEBPInfo.FrameCount := 1;
end;

function ReadWebpHeader(const Stream: TStream; var WEBPInfo: TWEBPInfo): Boolean;
var
  WH: TWebpHeader;
  WE: TWebpExtendedChunk;
  WL: TWebpLosslessChunk;
  WS: TWebpSimpleChunk;
  CW, CH: Integer;
  SizeBits: array [0..1] of Byte;
begin
  Result := False;
  Stream.Read(WH, SizeOf(WH));
  if (WH.RIFF <> 'RIFF') or (WH.WEBP <> 'WEBP') then
    Exit;

  if WH.Format = 'VP8X' then
  begin
    Stream.Read(WE, SizeOf(WE));

    WEBPInfo.Extended := True;
    WEBPInfo.ICC := BitTest(WE.Includes, 5);
    WEBPInfo.Alpha := BitTest(WE.Includes, 4);
    WEBPInfo.Exif := BitTest(WE.Includes, 3);
    WEBPInfo.XMP := BitTest(WE.Includes, 2);
    WEBPInfo.Animation := BitTest(WE.Includes, 1);
    WEBPInfo.FirstFrameAlpha := WEBPInfo.Alpha;

    CW := 0; CH := 0;
    Move(WE.CanvasWidth, CW, SizeOf(WE.CanvasWidth));
    Move(WE.CanvasHeight, CH, SizeOf(WE.CanvasHeight));
    CW := CW + 1; CH := CH + 1;

    WEBPInfo.Width := CW;
    WEBPInfo.Height := CH;

    Stream.Position := Stream.Position - (SizeOf(WE) - 4) + FixOdd(WE.ChunkSize);
    ReadNextChunks(Stream, WH.Size, WEBPInfo);
  end else if WH.Format = 'VP8L' then
  begin
    Stream.Read(WL, SizeOf(WL));
    CW := 0; CH := 0;

    Move(WL.Width, SizeBits, SizeOf(SizeBits));
    BitClear(SizeBits[1], 6);
    BitClear(SizeBits[1], 7);
    Move(SizeBits, CW, SizeOf(SizeBits));

    SizeBits[0] := 0;
    SizeBits[1] := 0;

    BitCopy(WL.Width[1], 6, SizeBits[0], 0);
    BitCopy(WL.Width[1], 7, SizeBits[0], 1);

    BitCopy(WL.Height[0], 0, SizeBits[0], 2);
    BitCopy(WL.Height[0], 1, SizeBits[0], 3);
    BitCopy(WL.Height[0], 2, SizeBits[0], 4);
    BitCopy(WL.Height[0], 3, SizeBits[0], 5);
    BitCopy(WL.Height[0], 4, SizeBits[0], 6);
    BitCopy(WL.Height[0], 5, SizeBits[0], 7);

    BitCopy(WL.Height[0], 6, SizeBits[1], 0);
    BitCopy(WL.Height[0], 7, SizeBits[1], 1);

    BitCopy(WL.Height[1], 0, SizeBits[1], 2);
    BitCopy(WL.Height[1], 1, SizeBits[1], 3);
    BitCopy(WL.Height[1], 2, SizeBits[1], 4);
    BitCopy(WL.Height[1], 3, SizeBits[1], 5);

    Move(SizeBits, CH, SizeOf(SizeBits));
    CW := CW + 1; CH := CH + 1;

    if BitTest(WL.Height[1], 4) then
    begin
      WEBPInfo.Alpha := True;
      WEBPInfo.FirstFrameAlpha := True;
    end;

    WEBPInfo.Compression := webpLossless;
    WEBPInfo.Width := CW;
    WEBPInfo.Height := CH;

    Stream.Position := Stream.Position - (SizeOf(WL) - 8);
    ReadNextChunks(Stream, WH.Size, WEBPInfo);

  end else if WH.Format = 'VP8 ' then
  begin
    Stream.Read(WS, SizeOf(WS));
    CW := 0; CH := 0;
    Move(WS.Width, CW, SizeOf(WS.Width));
    Move(WS.Height, CH, SizeOf(WS.Height));

    WEBPInfo.Compression := webpLossy;
    WEBPInfo.Width := CW;
    WEBPInfo.Height := CH;

    Stream.Position := Stream.Position - (SizeOf(WS) - 8);
    ReadNextChunks(Stream, WH.Size, WEBPInfo);
  end;

  Result := True;
  WEBPInfo.Valid := Result;
  if WEBPInfo.Valid then
  begin
    if WEBPInfo.FrameCount = 0 then
      WEBPInfo.FrameCount := 1;
  end;
end;

{ TWEBPInfo }

function TWEBPInfo.BitDepth: Integer;
begin
  if not Valid then
    Exit(0);
  if FirstFrameAlpha then
    Result := 32
  else
    Result := 24;
end;

end.
